﻿using System;
using System.Globalization;

namespace IC_DB
{
   /// <summary>
   /// Klasse mit Hilfsfunktionen für I²C-Übtragungsfolgen.
   /// </summary>
   public static class I2CHelpers
   {
      private enum State
      {
         Start,
         InComment,
         InDigit,
         InHexDegit
      }

      private enum CharType
      {
         NULL,
         Digit,
         HexDigit,
         X,
         Comma,
         Comment,
         NewSequence,
         EOF,
         NumberSign,
         Any
      }

      static CharType charType(char c)
      {
         if (c == '0')
            return CharType.NULL;
         if (c >= '1' && c <= '9')
            return CharType.Digit;
         if ((c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F'))
            return CharType.HexDigit;
         if (c == 'x' || c == 'X')
            return CharType.X;
         if (c == ';' || c == ',')
            return CharType.Comma;
         if (c == '\'')
            return CharType.Comment;
         if (c == '|')
            return CharType.NewSequence;
         if (c == '#')
            return CharType.NumberSign;
         if (c == '\0')
            return CharType.EOF;

         return CharType.Any;
      }

      private class Chars
      {
         char[] chars;
         public int pos = 0;
         public Chars(String x) { chars = x.ToCharArray(); }
         public char get
         {
            get
            {
               if (pos < chars.Length)
                  return chars[pos++];
               else
                  return '\0';
            }
         }
         public char peek
         {
            get
            {
               if (pos < chars.Length)
                  return chars[pos];
               else
                  return '\0';
            }
         }
         public void skipBlanks()
         {
            while ((pos < chars.Length) && (chars[pos] == ' '))
               pos++;
         }

      }

      /// <summary>
      /// Fehlercodes bei der Prüfung eines I2C-Xmit-Strings.
      /// </summary>
      public enum I2cErrorCode
      {
         /// <summary>Kein Fehler.</summary>
         None,
         /// <summary>Zeichen an dieser Posotion nicht erlaubt.</summary>
         InvalidAtPosition,
         /// <summary>Zeichen nicht erlaubt.</summary>
         InvalidChar,
         /// <summary>Angegebener Wert zu groß für ein Byte (>255).</summary>
         TooBig
      }


      /// <summary>
      /// Prüft den I²C-String. Falls fehlerfrei: liefert einen formatierten String zurück.
      /// </summary>
      /// <param name="i2c">Der zu überprüfende String.</param>
      /// <returns>Ein <see cref="Tuple{T1, T2, T3}"/>mit folgenden Inhalten:<para/>
      /// item1: wenn fehlerfrei: -1, bei Fehler: Position des ersten fehlerhaften Zeichen.
      /// Item2: Fehlercode bei Fehler
      /// item3: wenn fehlerfrei: Einen formatierten String, sonst "".
      /// item4: wenn fehlerfrei: Byte Arrays mit den zu übertregenden Daten, sonst <b>null</b>;
      /// </returns>
      public static Tuple<int, I2cErrorCode, String, I2cData?> CheckI2cDataStringSyntax(String i2c)
      {
         //TODO: Berücksichtigen, dass das letzte Byte die Anzahl einzulesender Zeichen sein könnte.
         State state = State.Start;
         State oldState = state;
         String formatedString = "";
         bool hasReadData = false; // bei der Analyse wurde festgestellt, dass I2C-Daten eingelesen werden sollen.

         int lastDigitPos = 0;

         I2cData? result = new I2cData();

         String currentDigitString = "";
         List<byte> currentBytes = new List<byte>();

         i2c = i2c.Replace(',', ';');
         i2c = i2c.Replace('X', 'x');
         i2c = i2c.Replace('"', '\'');

         Chars chars = new Chars(i2c);
         chars.skipBlanks();
         Char c = chars.get;
         while (c != '\0')
         {
            switch (state)
            {
               case State.Start: // Zeichen muss Ziffer oder x oder ' sein.
                  switch (charType(c))
                  {
                     case CharType.NULL: // So beginnt eine Dezimal oder Hexadezimalzahl
                        lastDigitPos = chars.pos;
                        var q = chars.peek;
                        if (charType(q) == CharType.X) // Wenn nächstes Zeichen x: Hexadezimalzahl.
                        {
                           c = chars.get; // x überspringen
                           c = chars.get; // erste Ziffer laden
                           state = State.InHexDegit;
                           formatedString += "0x";
                        }
                        else // Dezimalzahl
                        {
                           formatedString += c;
                           currentDigitString += c;
                           c = chars.get; // nächstes Zeichen
                           state = State.InDigit;
                        }
                        break;
                     case CharType.X: // So beginnt eine Hexadezimalzahl
                        lastDigitPos = chars.pos;
                        c = chars.get; // x überspringen
                        state = State.InHexDegit;
                        formatedString += "0x";
                        break;
                     case CharType.Digit: // So beginnt eine Dezimalzahl
                        lastDigitPos = chars.pos;
                        formatedString += c;
                        currentDigitString += c;
                        c = chars.get; // nächstes Zeichen
                        state = State.InDigit;
                        break;
                     case CharType.NumberSign: // So beginnt die Angabe der zu lesenden Zeichen
                        if (hasReadData)
                           return new Tuple<int, I2cErrorCode, string, I2cData?>(chars.pos - 1, I2cErrorCode.InvalidAtPosition, "", null); // Fehler
                        formatedString += " # ";
                        chars.skipBlanks();
                        c = chars.get; // nächstes Zeichen
                        hasReadData = true;
                        break;
                     case CharType.Comment: // So beginnt ein Kommentar
                        formatedString += " '";
                        c = chars.get;
                        oldState = state;
                        state = State.InComment;
                        break;
                     case CharType.NewSequence:
                        if (formatedString == "") // Am Anfang verboten
                           return new Tuple<int, I2cErrorCode, string, I2cData?>(chars.pos - 1, I2cErrorCode.InvalidAtPosition, "", null); // Fehler
                        result.Add(currentBytes.ToArray());
                        currentBytes = new List<byte>();
                        formatedString += " | ";
                        chars.skipBlanks();
                        c = chars.get;
                        break;
                     default:
                        return new Tuple<int, I2cErrorCode, string, I2cData?>(chars.pos - 1, I2cErrorCode.InvalidChar, "", null); // Fehler
                  }
                  break;
               case State.InComment: // Solange bis nächstes '.
                  if (charType(c) != CharType.Comment)
                  {
                     formatedString += c;
                     c = chars.get;
                  }
                  else
                  {
                     formatedString += "' ";
                     chars.skipBlanks();
                     c = chars.get;
                     state = oldState;
                  }
                  break;
               case State.InHexDegit:
                  var t = charType(c);
                  if (t == CharType.NULL || t == CharType.Digit || t == CharType.HexDigit)
                  {
                     string q = "" + c;
                     formatedString += q.ToUpper();
                     currentDigitString += c;
                     c = chars.get;
                  }
                  else // Ende der Hexadezimalzahl. Erlaubt sind hier ;'|# und Blank
                  {
                     int pos = chars.pos;

                     if (c == ' ')
                     {
                        chars.skipBlanks();
                        c = chars.get;
                     }
                     if (c == '\0') goto Done;

                     switch (charType(c))
                     {
                        case CharType.Comma:
                           formatedString += "; ";
                           chars.skipBlanks();
                           c = chars.get;
                           state = State.Start;

                           int b = Int32.Parse(currentDigitString, NumberStyles.AllowHexSpecifier, null);
                           if (b > 255)
                              return new Tuple<int, I2cErrorCode, string, I2cData?>(lastDigitPos - 1, I2cErrorCode.TooBig, "", null); // Fehler
                           currentBytes.Add((byte)b);
                           currentDigitString = "";
                           break;
                        case CharType.NewSequence:
                           b = Int32.Parse(currentDigitString, NumberStyles.AllowHexSpecifier, null);
                           if (b > 255)
                              return new Tuple<int, I2cErrorCode, string, I2cData?>(lastDigitPos - 1, I2cErrorCode.TooBig, "", null); // Fehler
                           currentBytes.Add((byte)b);
                           currentDigitString = "";
                           result.Add(currentBytes.ToArray());
                           currentBytes = new List<byte>();

                           formatedString += " | ";
                           chars.skipBlanks();
                           c = chars.get;
                           state = State.Start;
                           break;
                        case CharType.NumberSign:
                           if (hasReadData)
                              return new Tuple<int, I2cErrorCode, string, I2cData?>(chars.pos - 1, I2cErrorCode.InvalidAtPosition, "", null); // Fehler
                           hasReadData = true;
                           b = Int32.Parse(currentDigitString, NumberStyles.AllowHexSpecifier, null);
                           if (b > 255)
                              return new Tuple<int, I2cErrorCode, string, I2cData?>(lastDigitPos - 1, I2cErrorCode.TooBig, "", null); // Fehler
                           currentBytes.Add((byte)b);
                           currentDigitString = "";
                           result.Add(currentBytes.ToArray());
                           currentBytes = new List<byte>();

                           formatedString += " # ";
                           chars.skipBlanks();
                           c = chars.get;
                           state = State.Start;
                            break;
                        case CharType.Comment:
                           oldState = state;
                           formatedString += " \'";
                           state = State.InComment;
                           c = chars.get;
                           break;
                        default:
                           return new Tuple<int, I2cErrorCode, string, I2cData?>(chars.pos - 1, I2cErrorCode.InvalidChar, "", null); // Fehler
                     }
                  }
                  break;
               case State.InDigit:
                  t = charType(c);
                  if (t == CharType.NULL || t == CharType.Digit)
                  {
                     formatedString += c;
                     currentDigitString += c;
                     c = chars.get;
                  }
                  else // Ende der Dezimalzahl. Erlaubt sind hier ;'|# und Blank
                  {
                     int pos = chars.pos;

                     if (c == ' ')
                     {
                        chars.skipBlanks();
                        c = chars.get;
                     }
                     if (c == '\0') goto Done;

                     switch (charType(c))
                     {
                        case CharType.Comma:
                           formatedString += "; ";
                           chars.skipBlanks();
                           c = chars.get;
                           state = State.Start;

                           int b = Int32.Parse(currentDigitString);
                           if (b > 255)
                              return new Tuple<int, I2cErrorCode, string, I2cData?>(pos - currentDigitString.Length - 1, I2cErrorCode.TooBig, "", null); // Fehler
                           currentBytes.Add((byte)b);
                           currentDigitString = "";

                           break;
                        case CharType.NewSequence:
                           b = Int32.Parse(currentDigitString);
                           if (b > 255)
                              return new Tuple<int, I2cErrorCode, string, I2cData?>(lastDigitPos - 1, I2cErrorCode.TooBig, "", null); // Fehler
                           currentBytes.Add((byte)b);
                           currentDigitString = "";
                           result.Add(currentBytes.ToArray());
                           currentBytes = new List<byte>();

                           formatedString += " | ";
                           chars.skipBlanks();
                           c = chars.get;
                           state = State.Start;
                           break;
                        case CharType.NumberSign:
                           if (hasReadData)
                              return new Tuple<int, I2cErrorCode, string, I2cData?>(chars.pos - 1, I2cErrorCode.InvalidAtPosition, "", null); // Fehler
                           hasReadData = true;
                           b = Int32.Parse(currentDigitString);
                           if (b > 255)
                              return new Tuple<int, I2cErrorCode, string, I2cData?>(lastDigitPos - 1, I2cErrorCode.TooBig, "", null); // Fehler
                           currentBytes.Add((byte)b);
                           currentDigitString = "";
                           result.Add(currentBytes.ToArray());
                           currentBytes = new List<byte>();

                           formatedString += " # ";
                           chars.skipBlanks();
                           c = chars.get;
                           state = State.Start;
                           break;
                        case CharType.Comment:
                           oldState = state;
                           formatedString += " \'";
                           state = State.InComment;
                           c = chars.get;
                           break;
                        default:
                           return new Tuple<int, I2cErrorCode, string, I2cData?>(chars.pos - 1, I2cErrorCode.InvalidChar, "", null); // Fehler
                     }
                  }
                  break;
               default:
                  break;
            }
         }
      Done:
         if (currentDigitString != "")
         {
            if (state == State.InDigit)
            {
               int b = Int32.Parse(currentDigitString);
               if (b > 255)
                  return new Tuple<int, I2cErrorCode, string, I2cData?>(lastDigitPos - 1, I2cErrorCode.TooBig, "", null); // Fehler
               currentBytes.Add((byte)b);

            }
            if (state == State.InHexDegit)
            {
               int b = Int32.Parse(currentDigitString, NumberStyles.AllowHexSpecifier, null);
               if (b > 255)
                  return new Tuple<int, I2cErrorCode, string, I2cData?>(lastDigitPos - 1, I2cErrorCode.TooBig, "", null); // Fehler
               currentBytes.Add((byte)b);
            }
         }
         result.Add(currentBytes.ToArray());

         // Letzte Korrekturen an formatedString
         formatedString = formatedString.Replace(";  |", " |");
         formatedString = formatedString.Replace("|  ", "| ");
         formatedString = formatedString.Replace(";  #", " #");
         formatedString = formatedString.Replace("#  ", "# ");
         if (formatedString.EndsWith("; "))
            formatedString = formatedString.Substring(0, formatedString.Length - 2);

         if (hasReadData)
            return new Tuple<int, I2cErrorCode, string, I2cData?>(-2, I2cErrorCode.None, formatedString.Trim(), result);
         else
            return new Tuple<int, I2cErrorCode, string, I2cData?>(-1, I2cErrorCode.None, formatedString.Trim(), result);
      }
   }
}
