﻿using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Runtime.Serialization;
using System.Windows.Forms.VisualStyles;

namespace IC_DB
{
   /// <summary>
   /// Nimmt die Definitionen für ein IC auf.
   /// </summary>
   [DataContract]
   public class IcDefinition
   {
      /// <summary>Maximale Anzahl Pins. Ist vorgegegeben durch die Wahl der ZUF-Fassung.</summary>
      public const int MaxPins = 28;

      /// <summary>Interne eindeutige ID des Datensatzes.</summary>
      [DataMember]
      public Guid InternalID = Guid.NewGuid();
      /// <summary>Bezeichnung der Integrierten Schaltung, z.B. SN7400.</summary>
      [DataMember]
      public String Name = "";
      /// <summary>Beschreibung der Integrierten Schaltung, z.B.  Quad 2-Input NAND Gates.</summary>
      [DataMember]
      public String Description = "";
      /// <summary>Hinweise für die Testdurchführung.</summary>
      [DataMember]
      public String Notice = "";
      /// <summary>Kommentar zur Pin-Definition.</summary>
      [DataMember]
      public String PinsComment = "";
      /// <summary>Kommentar zur Test-Definition.</summary>
      [DataMember]
      public String TestComment = "";
      /// <summary>Beschreibung der einzelnen Pins.</summary>
      [DataMember]
      public PinDefinition[] Pins = new PinDefinition[MaxPins];
      /// <summary>Ein Set mit Testdaten.</summary>
      [DataMember]
      public TestSteps TestSet = new TestSteps();
      /// <summary>Schaltplan des IC.</summary>
      [DataMember]
      private Bitmap? circuitDiagram;
      /// <summary>Anzahl an Pins der Integrierten Schaltung.</summary>
      [DataMember]
      public int NumberOfPins = 14;

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="IcDefinition"/>-Klasse.
      /// </summary>
      /// <param name="Name">Bezeichnung des ICs (z.B. "SN74xx00").</param>
      /// <param name="Description">Beschreibung des ICs (z.B. "4-fach NAND-Gatter").</param>
      /// <param name="Manufacturer">Name des Herstellers. Wichtig für spezielle ICs.</param>
      public IcDefinition(String Name, String Description = "", String Manufacturer = "")
      {
         this.Name = Name;
         this.Description = Description;
         this.Notice = Manufacturer;
         for (int i = 0; i < MaxPins; i++)
         {
            Pins[i] = new PinDefinition();
         }
      }

      /// <summary>
      /// Ruft den Schaltplan ab oder legt ihn fest.
      /// </summary>
      public Bitmap? CircuitDiagram
      {
         get => circuitDiagram;
         set { circuitDiagram = value; }
      }

      /// <summary>
      /// Ermittelt, ob das IC eine I²C-Schnittstelle besitzt.
      /// </summary>
      /// <returns><b>true</b>, wenn das IC eine I²C-Schnittstelle besitzt.</returns>
      /// <remarks>Es mus genau ein Pin vom Typ <see cref="PinType.SDA"/> SDA und einer vom Typ <see cref="PinType.SCL"/> vorhanden sein.</remarks>
      public bool HasI2CPins()
      {
         int sdaPins = 0;
         int sckPins = 0;
         for (int i = 0; i < NumberOfPins - 1; i++)
         {
            if (Pins[i].Type == PinType.SDA) sdaPins++;
            if (Pins[i].Type == PinType.SCL) sckPins++;
         }
         return sdaPins == 1 && sckPins == 1;
      }

      /// <summary>
      /// Liefert eine tiefe Kopie des Objekts. Das kopierte Objekt erhält eine neue GUID.
      /// </summary>
      /// <returns>Eine tiefe Kopie des Objekts.</returns>
      public IcDefinition DeepCopy()
      {
         IcDefinition other = (IcDefinition)this.MemberwiseClone(); // Die elementaren Felder
         other.InternalID = Guid.NewGuid();
         other.Name = Name.DeepCopy();
         other.Description = Description.DeepCopy();
         other.Notice = Notice.DeepCopy();
         other.PinsComment = PinsComment.DeepCopy();
         other.TestComment = TestComment.DeepCopy();
         other.TestSet = TestSet.DeepCopy();
         other.Pins = new PinDefinition[MaxPins];
         for (int i = 0; i < Pins.Length; i++)
            other.Pins[i] = Pins[i].DeepCopy();
         if (CircuitDiagram != null)
            other.CircuitDiagram = new Bitmap(CircuitDiagram);
         return other;
      }

      /// <summary>
      /// Liefert die Bezeichnung des ICs.
      /// </summary>
      /// <returns>Die Bezeichnung des ICs.</returns>
      public override string ToString()
      {
         if (Description != "")
            return Name + ": " + Description;

         return Name;
      }

      /// <summary>
      /// Liefert die Bezeichnung des Pins.
      /// </summary>
      /// <returns>Die Bezeichnung des Pins.</returns>
      public String getPinsText()
      {
         String s = "";
         for (int i = 0; i < Pins.Length; i++)
         {
            s += "[" + i.ToString("00") + "]\t" + Pins[i].ToString() + "\r\n";
         }
         return s;
      }


      /// <summary>
      /// Prüft, ob die Feldinhalte (außer <see cref="InternalID"/>) identisch sind.
      /// </summary>
      /// <param name="ic">Das zu vergleichende <see cref="IcDefinition"/>-Objekt.</param>
      /// <returns><b>true</b>, wenn die Inhale identisch sind, ansonsnten <b>false</b>.</returns>
      public bool IsEqualTo(IcDefinition ic)
      {
         if (Name != ic.Name) return false;
         if (Description != ic.Description) return false;
         if (Notice != ic.Notice) return false;
         if (NumberOfPins != ic.NumberOfPins) return false;
         if (PinsComment != ic.PinsComment) return false;
         if (TestComment != ic.TestComment) return false;
         for (int i = 0; i < Pins.Length; i++)
            if (Pins[i] != ic.Pins[i]) return false;

         return CompareMemCmp(circuitDiagram, ic.circuitDiagram);
      }

      [DllImport("msvcrt.dll")]
      private static extern int memcmp(IntPtr b1, IntPtr b2, long count);

      /// <summary>
      /// Vergleicht zwei Bitmaps auf gleiche Größe und Inhalt.
      /// </summary>
      /// <param name="b1">Erste zu vergleichende Bitmap.</param>
      /// <param name="b2">Zweite zu vergleichende Bitmap.</param>
      /// <returns><b>true</b>, wenn die Inhalte identisch sind, ansonsten <b>false</b>.</returns>
      public static bool CompareMemCmp(Bitmap? b1, Bitmap? b2)
      {
         if (b1 == null && b2 == null) return true;
         if (b1 == null && b2 != null) return false;
         if (b1 != null && b2 == null) return false;
#pragma warning disable CS8602
         if (b1.Size != b2.Size) return false;
#pragma warning restore CS8602
         var bd1 = b1.LockBits(new Rectangle(new Point(0, 0), b1.Size), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
         var bd2 = b2.LockBits(new Rectangle(new Point(0, 0), b2.Size), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);

         try
         {
            IntPtr bd1scan0 = bd1.Scan0;
            IntPtr bd2scan0 = bd2.Scan0;

            int stride = bd1.Stride;
            int len = stride * b1.Height;

            return memcmp(bd1scan0, bd2scan0, len) == 0;
         }
         finally
         {
            b1.UnlockBits(bd1);
            b2.UnlockBits(bd2);
         }
      }


      /// <summary>
      /// Speichert diese IC-Definition in eine Datei.
      /// </summary>
      /// <param name="FileName">Name der Datei für diese IC-Definition.</param>
      public void SaveToFile(String FileName)
      {
         DataContractSerializer serializer = new DataContractSerializer(typeof(IcDefinition));
         using (FileStream fs = new FileStream(FileName, FileMode.Create))
         {
            serializer.WriteObject(fs, this);
         }
      }

      /// <summary>
      /// Liest die IC-Definition aus einer Datei.
      /// </summary>
      /// <param name="FileName">Name der Datei mit den IC-Definition.</param>
      /// <returns>Die eingelesene IC-Definitionen oder null bei Fehler.</returns>
      public static IcDefinition? ReadFromFile(String FileName)
      {
         IcDefinition? icDef;

         try
         {
            DataContractSerializer serializer = new DataContractSerializer(typeof(IcDefinition));
            using (FileStream fs = new FileStream(FileName, FileMode.Open))
            {
               icDef = (IcDefinition?)serializer.ReadObject(fs);
            }

            return icDef;
         }
         catch (Exception)
         {
            return null;
         }
      }

      Tuple<String, int, String> splitName(String Name)
      {
         String s = ""; // nimmt die Buchsteben zu Beginn auf
         String n = ""; // nimmt folgende Ziffern au
         String r = ""; // nimmt Rest auf
         int i = 0;


         while (i < Name.Length && Char.IsLetter(Name[i]))
         {
            s += Name[i++];
         }
         while (i < Name.Length && Char.IsDigit(Name[i]))
         {
            n += Name[i++];
         }
         while (i < Name.Length)
         {
            r += Name[i++];
         }
         int number = 0;
         int.TryParse(n,out number);
         return new Tuple<string, int, String>(s, number, r);

      }

      /// <summary>
      /// Namensvergleich zweier IC-Bezeichnung unter Berücksichtigung der Nummer
      /// </summary>
      /// <param name="other">Das <see cref="IcDefinition"/>-Objekt, mit dem verglichen werden soll.</param>
      /// <returns>-1: kleiner; 0: gleich; 1: größer</returns>
      public int NameCompare(IcDefinition other)
      {
         if (other.Name == Name)
            return 0;
         var myName = splitName(Name);
         var otherName = splitName(other.Name);
         var comparison = String.Compare(myName.Item1, otherName.Item1);
         if (comparison != 0)
            return comparison;
         if (myName.Item2 < otherName.Item2)
            return -1;
         if (myName.Item2 > otherName.Item2)
            return 1;
         return String.Compare(myName.Item3, otherName.Item3); ;
      }

   } // class
} // Namespace