﻿using System.Runtime.Serialization;

namespace IC_DB
{
   /// <summary>
   /// Datenbank für die IC-Definitionen.
   /// </summary>
   [DataContract]
   public class IcDefs
   {
      [DataMember]
      private List<IcDefinition> icDefs = new List<IcDefinition>();
      private bool dirty = false;

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="IcDefs"/>-Klasse.
      /// </summary>
      public IcDefs() { }


      /// <summary>
      /// Stellt die Methode dar, die das <see cref="DirtyChanged"/>-Ereignis der <see cref="IcDefs"/>-Klasse behandelt.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="e">Ein <see cref="IcSelectorDirtyChangedEventArgs"/>-Objekt, das die Ereignisdaten enthält.</param>
      public delegate void DirtyChangedEventHandler(object sender, IcSelectorDirtyChangedEventArgs e);

      /// <summary>Tritt ein, wenn ein Feld geändert wurde.</summary>
      public event DirtyChangedEventHandler? DirtyChanged;

      /// <summary>
      /// Setzt das Dirty-Flag asynchron.
      /// </summary>
      private void setDirty()
      {
         dirty = true;
         DirtyChanged?.Invoke(this, new IcSelectorDirtyChangedEventArgs(dirty));
      }

      /// <summary>
      /// Löscht das Dirty-Flag asynchron.
      /// </summary>
      private void resetDirty()
      {
         dirty = false;
         DirtyChanged?.Invoke(this, new IcSelectorDirtyChangedEventArgs(dirty));
      }

      /// <summary>
      /// Fügt der Liste der Definitionen eine neue hinzu.
      /// </summary>
      /// <param name="ic">Das <see cref="IcDefinition"/>-Objekt, das hinzugefügt werden soll.</param>
      /// <returns>Das eingefüge <see cref="IcDefinition"/>-Objekt.</returns>
      /// <exception cref="DuplicateIcDefinitionException">Ein Datensatz mit dem Namen ist bereits vorhanden.</exception>
      public IcDefinition Add(IcDefinition ic)
      {
         if (contains(ic.Name))
            throw new DuplicateIcDefinitionException("Bereits vorhanden " + ic.Name);
         icDefs.Add(ic);
         setDirty();
         return ic;
      }


      /// <summary>
      /// Fügt der Liste der Definitionen eine neue hinzu.
      /// </summary>
      /// <param name="name">Bezeichnung des ICs (z.B. "SN74xx00").</param>
      /// <param name="description">Beschreibung des ICs (z.B. "4-fach NAND-Gatter").</param>
      /// <param name="manufacturer">Name des Herstellers. Wichtig für spezielle ICs.</param>
      /// <returns>Das eingefüge <see cref="IcDefinition"/>-Objekt.</returns>
      /// <exception cref="DuplicateIcDefinitionException">Ein Datensatz mit dem Namen ist bereits vorhanden.</exception>
      public IcDefinition Add(String name, String description = "", String manufacturer = "")
      {
         if (contains(name))
            throw new DuplicateIcDefinitionException("Bereits vorhanden " + name);

         IcDefinition ic = new IcDefinition(name, description, manufacturer);
         icDefs.Add(ic);

         setDirty();
         return ic;
      }

      /// <summary>
      /// Entfernt das erste Vorkommen eines angegebenen Objekts aus der Liste.
      /// </summary>
      /// <param name="icDef">Das aus der Liste zu entfernende Objekt.</param>
      /// <exception cref="ArgumentException">Das Element konnte nicht entfernt werden. Nicht vorhanden?</exception>
      public void Remove(IcDefinition icDef)
      {
         if (!icDefs.Remove(icDef))
            throw new IcDefinitionNotFoundException(icDef.Name);
         setDirty();
      }

      /// <summary>
      /// Verändert das besthende Item.
      /// </summary>
      /// <param name="icDef">Das Objekt, dass geändert werden soll.</param>
      /// <returns>Das alte Objekt.</returns>
      /// <exception cref="ArgumentException">Das Element konnte nicht geändert werden.</exception>
      public IcDefinition Update(IcDefinition icDef)
      {
         foreach (var item in icDefs)
         {
            if (item.InternalID == icDef.InternalID)
            {
               Remove(item);
               if (contains(icDef.Name)) // Es gibt bereits ein IC mit dem geänderten Namen
               {
                  icDefs.Add(item); // Das alte Element wieder einfügen
                  throw new DuplicateIcDefinitionException(icDef.Name);
               }

               icDefs.Add(icDef);
               setDirty();
               return item;
            }
         }
         throw new IcDefinitionNotFoundException(icDef.Name);
      }


      /// <summary>
      /// Prüft, ob ein IC mit dem angegeben Namen in der Liste vorhanden ist.
      /// </summary>
      /// <param name="name">Due bezeichnung des zu suchenden Objekts. Groß/kleinschreibung wird ignoriert.</param>
      /// <returns><b>true</b>, wenn eine Definition mit dem angegeben Namen vorhanden ist, anderfalls <b>false</b>.</returns>
      private bool contains(String name)
      {
         foreach (IcDefinition item in icDefs)
         {
            if (item.Name.ToLower() == name.ToLower())
            {
               return true;
            }
         }

         return false;
      }

      /// <summary>
      /// Speichert die IC-Definitionen in eine Datei.
      /// </summary>
      /// <param name="FileName">Name der Datei für die IC-Definitionen.</param>
      public void SaveToFile(String FileName)
      {
         DataContractSerializer serializer = new DataContractSerializer(typeof(IcDefs));
         using (FileStream fs = new FileStream(FileName, FileMode.Create))
         {
            serializer.WriteObject(fs, this);
         }

         resetDirty();
      }

      /// <summary>
      /// Liest die IC-Definitionen aus einer Datei.
      /// </summary>
      /// <param name="FileName">Name der Datei mit den IC-Definitionen.</param>
      /// <returns>Die Liste der bekannteb IC-Definitionen.</returns>
      public static IcDefs ReadFromFile(String FileName)
      {
         IcDefs? newIcDefs;

         DataContractSerializer serializer = new DataContractSerializer(typeof(IcDefs));
         using (FileStream fs = new FileStream(FileName, FileMode.Open))
         {
            newIcDefs = (IcDefs?)serializer.ReadObject(fs);
         }

         if (newIcDefs == null) // Wenn nicht gelesen werden kann, wird eine Ausnahme geworfen.
            return new IcDefs(); // Dies ist ein Workaround für das nullable Objekt.
 
         return newIcDefs;
      }

      /// <summary>
      ///  liefert eine Readonly-Liste der vorhandenen Definitionen.
      /// </summary>
      /// <returns>Eine Readonly-Liste der vorhandenen Definitionen.</returns>
      public IReadOnlyCollection<IcDefinition> GetList()
      {
         return icDefs.AsReadOnly();
      }

      /// <summary>
      /// Gibt an, ob sich in der Liste Änderungen seit der letzten Speicherung vorgenommen wurden.
      /// </summary>
      /// <returns><b>true</b>, wenn die Liste ungespeicherte Änderunge enthält, ansonsten <b>false</b>.</returns>
      public bool IsDirty()
      {
         return dirty;
      }

      /// <summary>
      /// Zugriff auf ein einzelnes Element der Definitionen-Liste.
      /// </summary>
      /// <param name="index">Der nullbasierte Index des Elements.</param>
      /// <returns>Das selektierte Element.</returns>
      public IcDefinition this[int index]
      {
         get => icDefs[index];
      }
   }

   /// <summary>
   /// Ereignisargumente für das <see cref="IcDefs.DirtyChanged"/>-Ereignis.
   /// </summary>
   public class IcSelectorDirtyChangedEventArgs
   {
      /// <summary>
      /// Zustand des Dirty-Flags.
      /// </summary>
      public readonly bool dirty;

      /// <summary>
      /// Initialisert eine neue Instanz der <see cref="IcSelectorDirtyChangedEventArgs"/>-Klasse.
      /// </summary>
      /// <param name="dirty"></param>
      public IcSelectorDirtyChangedEventArgs(bool dirty)
      {
         this.dirty = dirty;
      }
   }
}
