﻿using System;
using System.Runtime.Serialization;

namespace IC_DB
{
   /// <summary>
   /// Klassifiziert den Testschritt.
   /// </summary>
   public enum TestStepType
   {
      /// <summary>Einfacher Ein/Ausgabe-Testschritt.</summary>
      /// <remarks>Input-Pins werden mit Werten belegt und die Output-Pins ausgelesen.</remarks>
      IO,
      /// <summary>Ein/Ausgabe-Testschritt mit Datenversand über I²C-Schnitstelle.</summary>
      /// <remarks>Input-Pins werden mit Werten belegt, dann die angegebenen Daten per I²C versendet und 
      /// anschließend die Output-Pins ausgelesen.</remarks>
      I2CXmit,

      /// <summary>Ein/Ausgabe-Testschritt mit Datenversand über I²C-Schnitstelle.</summary>
      /// <remarks>Input-Pins werden mit Werten belegt, dann die angegebenen Daten per I²C versendet und 
      /// anschließend über die I2C-Schnittstelle Daten eingelesen. Zum Schluss werden die Output-Pins ausgelesen.</remarks>
      I2CXmitAndRead

   }

   /// <summary>
   /// Repräsentiert einen Testschritt.
   /// </summary>
   [DataContractAttribute]
   public class TestStep
   {
      /// <summary>
      /// Der Typ des Testschritts.
      /// </summary>
      [DataMember]
      public TestStepType StepType = TestStepType.IO;

      /// <summary>Bezeichnung des Testschritts.</summary>
      [DataMember]
      public String Name = "";

      /// <summary>Array [0..(<see cref="IcDefinition.NumberOfPins"/>-1)] mit dem <see cref="LogicLevel"/> für jeden Pin.</summary>
      [DataMember]
      public LogicLevel[] Pins;

      /// <summary>
      /// String mit I²C Daten.
      /// </summary>
      /// <remarks>Einzelne Bytes durch ';' oder ',' getrennt.<para/>
      /// Dezimalzahlen zwischen 0 und 255.<para/>
      /// Hexadezimalzahlen beginnend mit 'x', 'X', '0x' oder '0X' zwischen 0x0 und 0xFF.<para/>
      /// Beginnend und endend mit einer Zahl.<para/>
      /// Leerzeichen sind erlaubt.<para/>
      /// Mehrere Transmit-Sequenzen durch "|" getrennt.<para/>
      /// Kommentare eingeschlossen in """ oder "'";<para/>
      /// Leere Einträge sind nicht erlaubt und werden entfernt: 
      /// ';' oder ',' zu Beginn oder am Ende oder direkt oder nur durch Leerzeichen getrennt aufeinderfolgend.
      /// </remarks>
      [DataMember]
      public String I2cXmitString = "";

      /// <summary>Die I2C-Daten als Liste von Byte-Arrays.</summary>
      [DataMember]
      public I2cData I2cXmitData = new I2cData();

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="TestStep"/>-Klasse.
      /// </summary>
      /// <param name="count">Anzahl Pins.</param>
      public TestStep(int count) { Pins = new LogicLevel[count]; }


      /// <summary>
      /// Liefert eine tiefe Kopie des Testschritts.
      /// </summary>
      /// <returns>Eine tiefe Kopie des Testschritts.</returns>
      public TestStep DeepCopy()
      {
         TestStep other = (TestStep)this.MemberwiseClone();
         other.Name = Name.DeepCopy();
         other.I2cXmitString = I2cXmitString.DeepCopy();
         other.I2cXmitData = I2cXmitData.DeepCopy();
         other.Pins = Pins.ToArray();
         return other;
      }

      /// <summary>
      /// Vergleicht, ob zwei Instanzen von <see cref="TestStep"/> den gleichen Inhalt haben.
      /// </summary>
      /// <param name="obj1">Erstes Vergleichsobjekt.</param>
      /// <param name="obj2">Zweites Vergleichsobjekt.</param>
      /// <returns><b>true</b>, wenn die Inhalte beider Instanzen gleich sind, ansonsten <b>false</b>.</returns>
      /// <remarks>Feld <see cref="I2cXmitData"/> wird nicht verglichen, maschinell generiert.</remarks>
      public static bool operator ==(TestStep obj1, TestStep obj2)
      {
         if (obj1.Name != obj2.Name)
            return false;

         if (obj1.StepType != obj2.StepType)
            return false;

         if (obj1.I2cXmitString != obj2.I2cXmitString)
            return false;

         if (obj1.Pins.Length != obj2.Pins.Length)
            return false;
         for (int i = 0; i < obj1.Pins.Length; i++)
            if (obj1.Pins[i] != obj2.Pins[i])
               return false;
         return true;
      }

      /// <summary>
      /// Vergleicht, ob zwei Instanzen von <see cref="TestStep"/> nicht den gleichen Inhalt haben.
      /// </summary>
      /// <param name="obj1">Erstes Vergleichsobjekt.</param>
      /// <param name="obj2">Zweites Vergleichsobjekt.</param>
      /// <returns><b>true</b>, wenn die Inhalte beider Instanzen nicht gleich sind, ansonsten <b>false</b>.</returns> 
      /// <remarks>Feld <see cref="I2cXmitData"/> wird nicht verglichen, maschinell generiert.</remarks>
      public static bool operator !=(TestStep obj1, TestStep obj2)
      { return !(obj1 == obj2); }

      /// <summary>
      /// Fungiert als Standardhashfunktion.
      /// </summary>
      /// <returns>Ein Hashcode für das aktuelle Objekt.</returns>
      public override int GetHashCode()
      {
         return Pins.GetHashCode();
      }

      /// <summary>
      /// Bestimmt, ob der aktuelle <see cref="TestStep"/> mit dem angegebenen Object übereinstimmt.
      /// </summary>
      /// <param name="obj">Das Vergleichsobjekt.</param>
      /// <returns><b>true</b>, wenn <paramref name="obj"/> vom Typ <see cref="TestStep"/> ist 
      /// und gleich dem aktuellen <see cref="TestStep"/> ist.</returns>
      public override bool Equals(object? obj)
      {
         if (ReferenceEquals(this, obj)) return true;

         var testStep = obj as TestStep;
         if (testStep is null) return false;

         return testStep == this;
      }

      /// <summary>
      /// Konvertiert den Wert dieser Instanz in die entsprechende Zeichenfolgendarstellung.
      /// </summary>
      /// <returns>Die Zeichenfolgendarstellung des Werts dieser Instanz.</returns>
      public override string ToString()
      {
         String s = "";
         for (int i = 0; i < Pins.Length; i++)
         {
            s+= "[" + i.ToString("00") + "]\t" + Pins[i].ToString() + "\r\n";
         }
         return s;
      }
   } // Klasse
} // namespac