﻿using IC_DB;

namespace IC_Tester_Controls
{
   /// <summary>
   /// Fenster zum Bearbeiten der Testschritte.
   /// </summary>
   public partial class TestEditor : Form
   {
      const int fixedColumns = 2; // Anzahl fester Spalten (Nr. & Name)
      IcDefinition ic;
      IcDefinition newIc; // Ic mit den geänderten Daten zum Vergleich

      bool hasI2CPins = false; // Zwischenspeicher für ic.HasI2C, Ermittlung ist aufwendig.

      /// <summary>
      /// <b>true</b>, wenn der Satz an Testschritten geändert wurde.
      /// </summary>
      public bool IsDirty = false;

      // Bei Closing
      // true, wenn Inhalte bereits in cmdSave_Click geprüft wurden. Das Fenster kann geschlossen werden.
      private bool isChecked = false;

      // Merker für die aktuell mit der Maus gewählte Zelle. Für Kontext-Menü.
      private DataGridViewCell? currentCell;

      // Zur Verwendung der Form im Readonly-Modus
      private ContextMenuStrip? StandardMenu = null;
      private ContextMenuStrip? I2cMenu = null;
      private ContextMenuStrip? XmitMenu = null;
      private Color InputColor = Color.LightBlue;
      private Color OutputColor = Color.LightCoral;
      private bool IsReadOnly = false;
      private bool CloseAllowed = true;

      #region [ Initialisierung ]
      /// <summary>
      /// Iniltialisiert eine neue Instanz der <see cref="TestEditor"/>-Klasse.
      /// </summary>
      /// <param name="ic">Das <see cref="IcDefinition"/>-Objekt, zu dem dieses Testschritte gehören.</param>
      /// <param name="readOnly"><b>false: Das Fenster wird nur angezeigt ohne Editiermöglichkeiten.</b></param>
      public TestEditor(IcDefinition ic, bool readOnly = false)
      {
         InitializeComponent();
         this.ic = ic;
         newIc = ic.DeepCopy();

         tbComment.Text = ic.TestComment;
         pbCircuit.Image = ic.CircuitDiagram;
         lblName.Text = ic.ToString();
         hasI2CPins = ic.HasI2CPins();
         IsReadOnly = readOnly;
         dataGridView1.ReadOnly = readOnly;
         if (!readOnly)
         {
            StandardMenu = menuStd;
            I2cMenu = menuI2c;
            XmitMenu = menuXmitData;
            if (!hasI2CPins)
            {
               cmdInsertI2CTransmit.Visible = false;
               toolStripSeparator1.Visible = false;
            }
         }
         else
         {
            OutputColor = Color.LightGray;
            lblOutputColor.BackColor = OutputColor;
            InputColor = Color.LightSteelBlue;
            lblInputColor.BackColor = InputColor;
            dataGridView1.DefaultCellStyle.SelectionBackColor = dataGridView1.DefaultCellStyle.BackColor;
            dataGridView1.DefaultCellStyle.SelectionForeColor = dataGridView1.DefaultCellStyle.ForeColor;
            dataGridView1.AllowUserToAddRows = false;
            dataGridView1.RowHeadersVisible = false;
            cmdSave.Visible = false;
            cmdAbort.Visible = false;
            CloseAllowed = false;
            Text = "Testschritte";
         }
         IsDirty = false;
      }

      private void TestEditor_Load(object sender, EventArgs e)
      {
         // Spalten anlegen
         // ------------------------------------
         dataGridView1.Columns.RemoveAt(fixedColumns); // 3. Spalte (Index = fixedColumns) ist Dummy-Pin-Spalte 
         dataGridView1.Columns[0].HeaderCell.Style.Alignment = DataGridViewContentAlignment.MiddleRight;
         for (int i = 0; i < ic.NumberOfPins; i++)
         {
            if (ic.Pins[i].Type.IsFunctionalPin())
            {
               var q = new DataGridViewTextBoxColumn();
               q.Name = "pin" + i + 1; // Array beginnt bei 0; Pin-Nummerierung bei 1
               q.HeaderText = "" + (i + 1) + ": " + ic.Pins[i].Name;
               q.Width = 70;
               q.Resizable = DataGridViewTriState.False;
               q.DefaultCellStyle.BackColor = Color.White;
               q.ReadOnly = true;
               q.DefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleCenter;
               q.HeaderCell.Style.Alignment = DataGridViewContentAlignment.MiddleCenter;
               q.SortMode = DataGridViewColumnSortMode.NotSortable;
               dataGridView1.Columns.Add(q);
            }
            else
            {
               var q = new DataGridViewComboBoxColumn();
               q.Name = "pin" + i + 1; // Array beginnt bei 0; Pin - Nummerierung bei 1
               q.HeaderText = "" + (i + 1) + ": " + ic.Pins[i].Name;
               q.Width = 70;
               q.Resizable = DataGridViewTriState.False;
               q.Items.AddRange(new object[] { "LOW", "HIGH", "IGNORE" });
               if (ic.Pins[i].Type == PinType.Input)
                  q.DefaultCellStyle.BackColor = InputColor;
               else if (ic.Pins[i].Type == PinType.Output || ic.Pins[i].Type == PinType.OpenCollector)
                  q.DefaultCellStyle.BackColor = OutputColor;
               else
               {
                  q.DefaultCellStyle.BackColor = Color.White;
                  q.ReadOnly = true;
               }
               q.FlatStyle = FlatStyle.Flat;
               q.HeaderCell.Style.Alignment = DataGridViewContentAlignment.MiddleCenter;
               q.SortMode = DataGridViewColumnSortMode.NotSortable;
               dataGridView1.Columns.Add(q);
            }
         }

         // Breite des Grids festlegen
         var totalWidth = dataGridView1.Columns.GetColumnsWidth(DataGridViewElementStates.None) + dataGridView1.RowHeadersWidth;
         dataGridView1.ClientSize = new Size(totalWidth + 2, dataGridView1.Height);
         dataGridView1.Refresh();
         Width = dataGridView1.Width + 40;
         var maxWidth = Screen.FromControl(this).Bounds.Width * 3 / 4;
         if (Width > maxWidth) Width = maxWidth;
         dataGridView1.Anchor |= AnchorStyles.Right;
         if (dataGridView1.Width > Width - 40)
            dataGridView1.Width = Width - 40;

         // Zeilen füllen
         // ------------------------
         dataGridView1.Rows.Clear();
         for (int step = 0; step < ic.TestSet.Count; step++)
         {
            var RowIndex = dataGridView1.Rows.Add();
            var testStep = ic.TestSet[step];
            dataGridView1.Rows[RowIndex].Tag = RowType.IoDef;

            dataGridView1.Rows[RowIndex].Cells[1].Value = testStep.Name;
            for (int pin = 0; pin < ic.NumberOfPins; pin++)
            {
               try
               {
                  if (ic.Pins[pin].Type.IsFunctionalPin())
                     dataGridView1.Rows[RowIndex].Cells[pin + fixedColumns].Value = ic.Pins[pin].Type.GetDescription();
                  else if (testStep.Pins[pin] == LogicLevel.Low)
                     dataGridView1.Rows[RowIndex].Cells[pin + fixedColumns].Value = "LOW";
                  else if (testStep.Pins[pin] == LogicLevel.High)
                     dataGridView1.Rows[RowIndex].Cells[pin + fixedColumns].Value = "HIGH";
                  else
                     dataGridView1.Rows[RowIndex].Cells[pin + fixedColumns].Value = "IGNORE";
               }
               catch (Exception)
               {
                  dataGridView1.Rows[RowIndex].Cells[pin + fixedColumns].Value = "IGNORE"; ;
               }
            }

            // I2C-Datenzeile anlegen
            if (testStep.StepType == TestStepType.I2CXmit || testStep.StepType == TestStepType.I2CXmitAndRead)
            {
               currentCell = dataGridView1.Rows[RowIndex].Cells[0];
               AddTransmitDataToRow(RowIndex);
               dataGridView1.Rows[RowIndex + 1].Cells[2].Value = testStep.I2cXmitString;
            }
         }

         // Schrittnummer update
         updateNumbering();
      }
      #endregion

      #region [ Zellen hinzufügen oder löschen ]

      // Contextmenü "Testschritt einfügen"
      private void cmdInsertStep_Click(object sender, EventArgs e)
      {
         if (currentCell == null)
            return;

         try
         {
            dataGridView1.Rows.Insert(currentCell.RowIndex);
            updateNumbering();
         }
         catch (Exception) { }
      }

      // Contextmenü "I²C-Transmit einfügen".
      private void cmdInsertI2CTransmit_Click(object sender, EventArgs e)
      {
         // https://stackoverflow.com/questions/7055918/datagridview-cells-of-one-column-cant-have-different-type
         if (currentCell == null)
            return;
         AddTransmitDataToRow(currentCell.RowIndex);
      }

      /// <summary>
      /// TransmitData an Row anfügen.
      /// </summary>
      /// <param name="RowIndex">Row, an die angefügt werden soll.</param>
      private void AddTransmitDataToRow(int RowIndex)
      {
         int ri;
         try
         {
            ri = RowIndex + 1; // ri ist der Index der neuen Row.
            dataGridView1.Rows.Insert(ri); // Hinter die aktuelle Row einfügen.
         }
         catch (Exception)
         {
            return;
         }

         dataGridView1.Rows[ri].Tag = RowType.I2cXmit;
         dataGridView1.Rows[ri].ContextMenuStrip = XmitMenu;

         dataGridView1[1, ri].Value = "I²C Xmit";
         for (int i = 2; i < dataGridView1.Rows[ri].Cells.Count; i++)
         {
            DataGridViewTextBoxCell TextBoxCell = new DataGridViewTextBoxCellEx();
            dataGridView1[i, ri] = TextBoxCell;
            dataGridView1[i, ri].Style.BackColor = Color.White;
         }

         ((DataGridViewTextBoxCellEx)dataGridView1[2, ri]).ColumnSpan = dataGridView1.Rows[ri].Cells.Count - fixedColumns;
         dataGridView1[2, ri].Value = "I²C-Daten eingeben";
         dataGridView1[2, ri].ReadOnly = false;
         dataGridView1[1, ri].ReadOnly = true;
         dataGridView1[0, ri].Value = "";
      }

      private void cmdDuplicate_Click(object sender, EventArgs e)
      {
         if (currentCell == null)
            return;

         try
         {
            dataGridView1.Rows.Insert(currentCell.RowIndex);

            var newRow = dataGridView1.Rows[currentCell.RowIndex - 1];
            var oldRow = dataGridView1.Rows[currentCell.RowIndex];
            for (int i = 0; i < newRow.Cells.Count; i++)
               newRow.Cells[i].Value = oldRow.Cells[i].Value;

            if (dataGridView1.Rows[currentCell.RowIndex + 1].getRowType() == RowType.I2cXmit)
            {
               var temp = dataGridView1.Rows[currentCell.RowIndex + 1].Cells[2].Value;
               AddTransmitDataToRow(currentCell.RowIndex - 1);
               dataGridView1.Rows[currentCell.RowIndex - 1].Cells[2].Value = temp;
            }
         }
         catch (Exception) { }

         updateNumbering();
      }

      // Contextmenü "Testschritt löschen".
      private void cmdRemoveStep_Click(object sender, EventArgs e)
      {
         if (currentCell == null)
            return;
         if (MessageBox.Show("Daten löschen?", "IC-Tester", MessageBoxButtons.OKCancel, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.OK)
            try
            {
               dataGridView1.Rows.RemoveAt(currentCell.RowIndex);
               updateNumbering();
            }
            catch (Exception) { }
      }

      // Ereignis: neue Zeile hinzugefügt: Funktionale Zellen vorbelegen.
      private void dataGridView1_RowsAdded(object sender, DataGridViewRowsAddedEventArgs e)
      {
         dataGridView1.Rows[e.RowIndex].Tag = RowType.IoDef;

         // Zellen für funktionale Pins beschriften.
         for (int i = 0; i < ic.NumberOfPins; i++)
         {
            if (ic.Pins[i].Type.IsFunctionalPin())
               dataGridView1.Rows[e.RowIndex].Cells[i + fixedColumns].Value = ic.Pins[i].Type.GetDescription();
         }
         if (hasI2CPins)
            dataGridView1.Rows[e.RowIndex].ContextMenuStrip = StandardMenu;
         else
            dataGridView1.Rows[e.RowIndex].ContextMenuStrip = I2cMenu;
         updateNumbering();
      }

      // Nummerrierung der Testschritte überarbeiten.
      private void updateNumbering()
      {
         int cnt = 1;
         for (int i = 0; i < dataGridView1.Rows.Count; i++)
         {
            var rt = dataGridView1.Rows[i].getRowType();
            if (rt == RowType.IoDef)
               dataGridView1.Rows[i].Cells[0].Value = cnt++;
            else
               dataGridView1.Rows[i].Cells[0].Value = (cnt-1).ToString() + "x";
         }
      }
      #endregion

      #region [ Besonders Verhalten des DataGridView ]
      // DropDown der ComboBox bei Klick in den Inhalt.
      private void dataGridView1_CellMouseDown(object sender, DataGridViewCellMouseEventArgs e)
      {
         if (e.ColumnIndex < 0)
            return;
         if (e.RowIndex < 0)
            return;
         if (IsReadOnly)
            return;
         try
         {
            currentCell = dataGridView1.Rows[e.RowIndex].Cells[e.ColumnIndex];
            if (e.Button == MouseButtons.Left)
            {
               var cell = dataGridView1.Rows[e.RowIndex].Cells[e.ColumnIndex];
               if (cell != null)
                  if (cell.GetType() == typeof(DataGridViewComboBoxCell))
                  {
                     DataGridViewComboBoxCell c = (DataGridViewComboBoxCell)cell;
                     {
                        dataGridView1.CurrentCell = cell;
                        dataGridView1.BeginEdit(true);
                        DataGridViewComboBoxEditingControl editingControl = (DataGridViewComboBoxEditingControl)dataGridView1.EditingControl;
                        editingControl.DroppedDown = true;
                     }
                  }
            }
         }
         catch (Exception)
         {
         }

      }

      // Speichert die aktuell ausgewählte Zelle in Feld 'currentCell' zwischen
      private void dataGridView1_CellEnter(object sender, DataGridViewCellEventArgs e)
      {
         if (e.ColumnIndex < 0)
            return;
         if (e.RowIndex < 0)
            return;
         currentCell = dataGridView1.Rows[e.RowIndex].Cells[e.ColumnIndex];
      }
      #endregion

      #region [ Inhalte Prüfen ]
      /// <summary>
      /// Legt ein neues <see cref="TestStep"/>-Objekt an und 
      /// ermittelt die Felder <see cref="TestStep.Name"/> und <see cref="TestStep.Pins"/>
      /// aus einer <see cref="DataGridViewRow"/> vom Typ <b>IO</b>.
      /// </summary>
      /// <param name="row">Die <see cref="DataGridViewRow"/>, die auisgelesen werden soll.</param>
      /// <returns>Das ermittelte <see cref="TestStep"/>-Objekt.</returns>
      /// <remarks>Ggf. in der Folgezeile vorhandene I2C-Daten werden nicht ausgelesen.</remarks>
      TestStep getTestStepIoDataFromDataGridRow(DataGridViewRow row)
      {
         TestStep ts = new TestStep(ic.NumberOfPins);
         if (row.Cells[1].Value == null)
            ts.Name = "";
         else
            ts.Name = (String)row.Cells[1].Value;

         for (int col = fixedColumns; col < ic.NumberOfPins + fixedColumns; col++)
         {
            if (row.Cells[col].Value is null)
               ts.Pins[col - fixedColumns] = LogicLevel.Ignore;
            else if (row.Cells[col].Value.Equals("LOW"))
            {
               ts.Pins[col - fixedColumns] = LogicLevel.Low;
            }
            else if (row.Cells[col].Value.Equals("HIGH"))
            {
               ts.Pins[col - fixedColumns] = LogicLevel.High;
            }
            else
               ts.Pins[col - fixedColumns] = LogicLevel.Ignore;
         }
         return ts;
      }

      /// <summary>
      /// Prüft, ob der in der angegeben Zelle erfasste I2C-Xmit-String den Regeln entspricht. 
      /// Markiert im Fehlerfall die fehlerhafte Stelle.
      /// </summary>
      /// <param name="cell">Die zu püfende Zelle.</param>
      /// <param name="ts"><see cref="TestStep"/>-Objekt zur Aufnahme der I2cXmitData.</param>
      /// <returns><b>true</b>, wenn der I2C-Xmit-String den Regeln entspricht, ansonsten <b>false</b>.</returns>
      private bool CheckI2cDataAndMarkErrors(DataGridViewCell cell, TestStep ts)
      {
         var q = cell.Value;
         if (q == null)
            throw new NullReferenceException("CheckI2cDataMarkErrors");


         String s = (String)q;
         var result = I2CHelpers.CheckI2cDataStringSyntax(s);
         if (result.Item1 > -1) // Fehler in der Zeile
         {
            dataGridView1.CurrentCell = cell;
            dataGridView1.BeginEdit(true);
            var m = (DataGridViewTextBoxEditingControl)dataGridView1.EditingControl;
            m.SelectionStart = result.Item1;
            m.SelectionLength = 1;
            Console.Beep();

            var dgvPos = dataGridView1.GetCellDisplayRectangle(2, cell.RowIndex, false).Location;

            lblError.Location = dataGridView1.Location;
            lblError.Left += dgvPos.X;
            lblError.Top += dgvPos.Y - 30;
            switch (result.Item2)
            {
               case I2CHelpers.I2cErrorCode.None:
                  break;
               case I2CHelpers.I2cErrorCode.InvalidAtPosition:
                  lblError.Text = "An dieser Position nicht erlaubt";
                  break;
               case I2CHelpers.I2cErrorCode.InvalidChar:
                  lblError.Text = "Zeichen nicht erlaubt";
                  break;
               case I2CHelpers.I2cErrorCode.TooBig:
                  lblError.Text = "Zahlenwert größer als 255 (0xFF)";
                  break;
               default:
                  break;
            }
            lblError.Visible = true;
            timer1.Enabled = true; // Timer löscht die Fehlermeldung wieder.
            return false;
         }
         else
         {
            cell.Value = result.Item3;
            ts.I2cXmitString = result.Item3;
            if (result.Item4 == null)
               throw new NullReferenceException("CheckI2cDataAndMarkErrors");
            ts.I2cXmitData = result.Item4;
            if (result.Item1 == -2)
               ts.StepType = TestStepType.I2CXmitAndRead;
            return true;
         }
      }

      /// <summary>
      /// Erstellt ein <see cref="TestSteps"/>-Objekt aus dem Datagrid.
      /// </summary>
      /// <param name="checkAndIncludeI2cXmitData">Die I2C-Daten (<see cref="TestStep.I2cXmitData"/>) sollen 
      /// ebenfalls geprüft und ermittelt werden.</param>
      /// <see cref="TestStep.I2cXmitString"/> und <see cref="TestStep.StepType"/> werden auf jeden Fall ermittelt.
      /// <returns>Das aus dem Datagrid erstellte <see cref="TestSteps"/>-Objekt oder null, wenn bei der Überprüfung
      /// von des I2cXmitString ein Fehler auftrat. Die entsprechende Stelle ist dann markiert.</returns>
      private TestSteps? MakeTestSet(bool checkAndIncludeI2cXmitData)
      {
         TestSteps testSet = new TestSteps();

         for (int rowIndex = 0; rowIndex < dataGridView1.Rows.Count - 1; rowIndex++)
         {
            RowType rt = dataGridView1.Rows[rowIndex].getRowType();

            if (rt == RowType.IoDef)
            {
               var row = dataGridView1.Rows[rowIndex];
               var ts = getTestStepIoDataFromDataGridRow(row); // Name und Pins ermitteln, StepType default.

               // I2cXmitString ermitteln und StepType anpassen
               try // try: falls auf die nächste Zeile nicht zugegriffen werden kann.
               {
                  rt = dataGridView1.Rows[rowIndex + 1].getRowType();
                  if (rt == RowType.I2cXmit)
                  {
                     var v = dataGridView1.Rows[rowIndex + 1].Cells[2].Value;
                     if (v != null)
                        ts.I2cXmitString = (String)v;
                     ts.StepType = TestStepType.I2CXmit;
                     rowIndex++;

                     if (checkAndIncludeI2cXmitData)
                        if (!CheckI2cDataAndMarkErrors(dataGridView1.Rows[rowIndex].Cells[2], ts)) // XmitString prüfen und Fehler markieren
                           return null; // Fehler
                  }
               }
               catch (Exception) { }

               testSet.Add(ts);
            }
         }
         return testSet;
      }
      #endregion

      #region [ Bearbeitung beenden ]


      private void cmdSave_Click(object sender, EventArgs e)
      {
         var testSet = MakeTestSet(checkAndIncludeI2cXmitData: true);
         if (testSet is null)
            return;

         if (testSet != ic.TestSet)
         {
            ic.TestSet = testSet;
            IsDirty = true;
         }

         if (tbComment.Text != ic.TestComment)
         {
            ic.TestComment = tbComment.Text;
            IsDirty = true;
         }
         isChecked = true;
         Close();
      }

      private void cmdAbort_Click(object sender, EventArgs e)
      { // Prüfung auf Veränderungen erfolgt in estEditor_FormClosing.
         Close();
      }

      private void TestEditor_FormClosing(object sender, FormClosingEventArgs e)
      {
         if (isChecked) // Form kann geschlossen werden
            return;

         if (!CloseAllowed)
         {
            e.Cancel = true;
            return;
         }

         if (IsReadOnly)
            return;

         // Gab es Änderungen?
         var testSet = MakeTestSet(checkAndIncludeI2cXmitData: false);
         if (testSet is null) // Ist nur null, wenn auch Syntaxprüfung stattfindet.
            throw new NullReferenceException("TestEditor_FormClosing");

         if (testSet != ic.TestSet || tbComment.Text != ic.TestComment)
         {
            var rtc = MessageBox.Show("Änderungern verwerfen?", "IC-Tester", MessageBoxButtons.OKCancel, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2);
            if (rtc != DialogResult.OK)
            {
               e.Cancel = true;
               return;
            }
         }
         IsDirty = false;
      }
      #endregion

      // Löscht die Fehlermeldung wieder.
      private void timer1_Tick(object sender, EventArgs e)
      {
         timer1.Enabled = false;
         lblError.Visible = false;
      }

      /// <summary>
      /// Markiert die Fehlerhaften Pins.
      /// </summary>
      /// <param name="TestStepNo">Testschrittnummer</param>
      /// <param name="ErrorPins">Liste der Pins mit fehlerhaften Zustand.</param>
      /// <param name="isI2cError">Bei dem Faherl handelt es sich um einen Fehler beim Test der I²C-Schnitstelle.</param>
      public void MarkErrors(int TestStepNo, List<int> ErrorPins, bool isI2cError)
      {
         for (int i = 0; i < dataGridView1.Rows.Count; i++)
         {
            var row = dataGridView1.Rows[i];
            if (row.Cells.Count > 0 && row.Cells[0] != null)
            {
               bool isI2cRow = false;
               string cell0String = "";
               string? cell0Value = row.Cells[0].Value.ToString();
               if (cell0Value != null)
                  cell0String = cell0Value.ToString();

               if (cell0String.EndsWith("x"))
               {
                  isI2cRow = true;
                  cell0String =
                  cell0String = cell0String.Replace("x", "");
               }

               int testStepNumber = int.Parse(cell0String);
               if (testStepNumber == TestStepNo)
               {
                  if (isI2cError && isI2cRow)
                     row.Cells[2].Style.BackColor = Color.Red;
                  if (!isI2cError && !isI2cRow)
                     foreach (int pin in ErrorPins)
                        row.Cells[pin + 2].Style.BackColor = Color.Red;
               }
            }
         }
      }

      /// <summary>
      /// Löscht die Fehlermarkierungen.
      /// </summary>
      public void ResetErrorMarks()
      {
         for (int i = 0; i < dataGridView1.Rows.Count; i++)
         {
            var row = dataGridView1.Rows[i];
            for (int c = 0; c < row.Cells.Count; c++)
               if (row.Cells[c].Style.BackColor == Color.Red)
                  row.Cells[c].Style.BackColor = OutputColor;
         }
      }

      /// <summary>
      /// Schließt das Fenster im ReadOnly- (Anzeige-) Modus.
      /// </summary>
      public void CloseForm()
      {
         CloseAllowed = true;
         Close();
      }


      private const int CP_NOCLOSE_BUTTON = 0x200;
      /// <summary>
      /// Deaktiviert den Close-Button (X).
      /// </summary>
      protected override CreateParams CreateParams
      {
         get
         {
            CreateParams myCp = base.CreateParams;
            myCp.ClassStyle = myCp.ClassStyle | CP_NOCLOSE_BUTTON;
            return myCp;
         }
      }
   } // class
} // namespace