﻿using System.Diagnostics.CodeAnalysis;
using System.Reflection;

namespace IC_Tester_Controls
{
   /// <summary>
   /// Legt einen tool tip text für ein Element fest.
   /// </summary>
   [AttributeUsage(AttributeTargets.Field)]
   public class ToolTipAttribute : Attribute
   {
      /// <summary>
      /// Legt die Voreinstellung für das <see cref='ToolTipAttribute'/> fest (ein leerer String).
      /// </summary>
      public static readonly ToolTipAttribute Default = new ToolTipAttribute();

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="ToolTipAttribute"/>-Klasse.
      /// </summary>
      public ToolTipAttribute() : this(string.Empty)
      {
      }

      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="ToolTipAttribute"/>-Klasse.
      /// </summary>
      /// <param name="text">Der anzuzeigende Text.</param>
      public ToolTipAttribute(string text)
      {
         ToolTipValue = text;
      }

      /// <summary>
      /// Ruft den in diesem Attribut gespeicherten ToolTip-Text ab.
      /// </summary>
      public virtual string ToolTip => ToolTipValue;

      /// <summary>
      /// Lese-/Schreibeigenschaft, die die im QuickInfo-Text gespeicherte Zeichenfolge direkt ändert.
      /// </summary>
      protected string ToolTipValue { get; set; }

      /// <inheritdoc/>
      public override bool Equals([NotNullWhen(true)] object? obj) =>
          obj is ToolTipAttribute other && other.ToolTip == ToolTip;

      /// <inheritdoc/>
      public override int GetHashCode() => ToolTip?.GetHashCode() ?? 0;

      /// <inheritdoc/>
      public override bool IsDefaultAttribute() => Equals(Default);



      /// <summary>
      /// Liest das <see cref="ToolTipAttribute"/> des Enumerationswertes aus.
      /// </summary>
      /// <param name="value">Der Enumerationswerte, dessen <see cref="ToolTipAttribute"/> ermittelt werden soll.</param>
      /// <returns>Der beim Enumerationswert hintelegte Beschreibungstext oder der Name des Enumerationswerts.</returns>
      public static string GetToolTipText<T>(T value) where T : Enum
      {
         FieldInfo? fi = value.GetType().GetField(value.ToString());

         if (fi != null)
         {
            ToolTipAttribute[] attributes = (ToolTipAttribute[])fi.GetCustomAttributes(typeof(ToolTipAttribute), false);
            if (attributes != null && attributes.Length > 0)
            {
               return attributes[0].ToolTip;
            }
         }
         return value.ToString();
      }
   }
}
