﻿using System.Management;
using static UrsComPortExtensions.ComPortSelector;

namespace UrsComPortExtensions
{
   /// <summary>
   /// Beschreibung eines ComPorts.
   /// </summary>
   public class ComPortDescriptor
   {
      /// <summary>Name des Ports</summary>
      /// <example><span style="color: red"><b>'COM24'</b></span></example>
      public readonly string PortName;

      /// <summary>Beschreibung des Ports</summary>
      /// <example><span style="color: red"><b>'Prolific USB-to-Serial Comm Port'</b></span></example>
      public readonly string Description;

      /// <summary>Nummer des Ports</summary>
      /// <example><span style="color: red"><b>'24'</b></span></example>
      public readonly int PortNumber;

      /// <summary>Eindeutige ID des Geräts.</summary>
      public readonly String DeviceID;

      /// <summary>Seriennummer des Geräts.</summary>
      public readonly String SerialNo;


      /// <summary>Bezeichnung, die angezeigt wwerden soll.
      /// 
      /// </summary>
      public String DisplayName;

      /// <summary>
      /// Erstellt eine neue Instanz der <see cref="ComPortDescriptor"/>-Klasse.
      /// </summary>
      /// <param name="Portnumber">Nummer des Ports, z.B. 24.</param>
      /// <param name="PortName">Name des Ports, z.B. 'COM24'.</param>
      /// <param name="Description">Beschreibung des Ports, z.B. 'Prolific USB-to-Serial Comm Port'.</param>
      /// <param name="DeviceID">Eindeutige ID des Geräts, z.B. 'USB\VID_067B&amp;PID_2303\6&amp;F9BED0F&amp;0&amp;4'.</param>
      private ComPortDescriptor(int Portnumber, string PortName, string Description, String DeviceID)
      {
         PortNumber = Portnumber;
         this.PortName = PortName;
         this.Description = Description;
         this.DeviceID = DeviceID;
         DisplayName = Description + " (" + PortName + ")";

         var parts = DeviceID.Split('\\');
         if (parts.Length > 2)
            SerialNo = parts[2];
         else
            SerialNo = "";
      }


      /// <summary>
      /// Liefert den Port-Namen, wie er zum Öffnen für native Window-Funktionen benötigt wird.
      /// </summary>
      /// <returns>Der Port-Name, wie er zum Öffnen benötigt wird.</returns>
      /// <remarks>Ports &gt; 9 benötigen den Prefix '//./'.</remarks>
      public string ToFileName()
      {
         if (PortNumber > 9)
         {
            return "//./" + PortName;
         }
         else
         {
            return PortName;
         }
      }

      /// <summary>
      /// Liefert eine anzeigefähge Beschreibung des ComPorts.
      /// </summary>
      /// <returns>Eine anzeigefähge Beschreibung des ComPorts.</returns>
      /// <example>Prolific USB-to-Serial Comm Port (COM24)</example>
      public override string ToString()
      {
         return DisplayName;
      }


      /// <summary>
      /// Liefert eine Liste aller aktiven ComPorts.
      /// </summary>
      /// <param name="NameFilter">Die Portbezeichnung muss den angegebenen Text enthalten. Leerstring: Es werden alle Comports ausgegegeb.</param>
      /// <returns>Eine Liste aller aktiven ComPorts.</returns>
      /// <remarks>Die Ermittlung erfolgt über die WMI. Das NUGET-Paket <b>System.Management</b> muss eingebunden sein.</remarks>
      public static List<ComPortDescriptor> GetPorts(string NameFilter = "")
      {
         return GetPorts(NameFilter, null);
      }

      /// <summary>
      /// Liefert eine Liste aller aktiven ComPorts.
      /// </summary>
      /// <param name="NameFilter">Die Portbezeichnung muss den angegebenen Text enthalten. Leerstring: Es werden alle Comports ausgegegeb.</param>
      /// <param name="comPortFoundEventHandler">Damit die statische Funktion das Ereignis an die richtige Callback-Funktion sendet.</param>
      /// <returns>Eine Liste aller aktiven ComPorts.</returns>
      /// <remarks>Die Ermittlung erfolgt über die WMI. Das NUGET-Paket <b>System.Management</b> muss eingebunden sein.</remarks>
      internal static List<ComPortDescriptor> GetPorts(string NameFilter, ComPortFoundEventHandler? comPortFoundEventHandler)
      {
         var ports = new List<ComPortDescriptor>();
         ManagementObjectCollection pReturn;
         ManagementObjectSearcher pSearch;

         if (NameFilter is null)
         {
            NameFilter = "";
         }

         pSearch = new ManagementObjectSearcher("Select * from Win32_PnPEntity");
         pReturn = pSearch.Get();

         foreach (var pObject in pReturn)
         {
            object? propertyValue = pObject.GetPropertyValue("Name");
            if (propertyValue == null)
               continue;

            String Name = (String)propertyValue;

            if (!(Name.Contains("(COM") || Name.ToLower().Contains("modem"))) // ComPorts selektieren
               continue;

            if (Name.Contains("COM#")) // Com0Com liefer einen zusätlichen Eintrag "COM# <-> COM#".
               continue;

            if (NameFilter != "") // Filter berücksichtigen.
               if (!Name.Contains(NameFilter))
                  continue;

            try
            {
               if (!Name.Contains("COM")) // für "modem"
                  Name += " (" + pObject.GetPropertyValue("DeviceID") + ")";

               // Bezeichnung und Nummer aus dem Namen ermitteln.
               String tmp = Name;
               int posCom = Name.IndexOf(" (COM");
               Name = tmp.Substring(posCom + 2);
               Name = Name.Substring(0, Name.Length - 1);
               int Number = int.Parse(Name.Substring(3));
               tmp = tmp.Substring(0, posCom);

               var id = pObject.GetPropertyValue("DeviceID");
               if (id == null) id = "";

               var cpd = new ComPortDescriptor(Number, Name, tmp, (String)id);

               if (comPortFoundEventHandler != null)
               {
                  ComPortFoundEventArgs cpfev = new ComPortFoundEventArgs(cpd);
                  comPortFoundEventHandler?.Invoke(new object(), cpfev);
                  if (cpfev.AddDeviceToList)
                     ports.Add(cpd);
               }
               else
                  ports.Add(cpd);
            }
            catch (Exception)
            {
            }
         }

         return ports.OrderBy(o => o.PortNumber).ToList(); ;
      }
   }
}